//
// by melanion.info
//
// TSimpleGate is for FMX apps
//
// FMX.WebBrowser.TWebBrowser.OnShouldStartLoadWithRequest
//     Sydney: not fire on MS-Windows
//     Alexandria: OK on all platforms
//

//
// ** TSimpleGate usage
//
// <script language="javascript" type="text/javascript">
//		function OnButtonClick() {                               // ex. when a event occurrs
//      var s = 'button clicked !';
//			window.location.href = 'delphi:MyProc(' + s + ')';     // Javascript calls a delphi function
//		}

//    function callFromDelphi() {                              // when Delphi calls a Javascript function
//      var resultValue =  ...
//      window.location.href = 'delphi:return=' + resultValue; // returns resultValue to Delphi
//    }
// </script>
//
// <body>
//   <a href="delphi:MyProc(OK)">call delphi</a>
//

unit SimpleGateUnit;

interface

uses
  System.Classes, System.SysUtils, FMX.Types,
  IdURI, System.DateUtils, System.Rtti, FMX.WebBrowser, System.Messaging;

type
  TResultProc = reference to procedure (const iResult: String);

  TSimpleGate = class
  strict private
    FReceiver: TObject;
    FWebBrowser: TWebBrowser;
    FScheme: String;
    FTag: integer;
    FResultProc: TResultProc;
    FLastReturnTime: TDateTime;
    FLastReturnString: string;
    FScriptLoaded: boolean;
    function CheckURL(var AText: string): boolean;
    procedure WebBrowserOnShouldStartLoadWithRequest(ASender: TObject; const URL: string);
    procedure CallEvent(AText: string);
    procedure MessageReceiver(const Sender: TObject; const M: TMessage);

  public
    constructor Create(const Receiver: TObject; const WebBrowser: TWebBrowser; const Scheme: String; Tag: integer);
    procedure CallScript(const Script: String; const ResultProc: TResultProc);
    property Loaded: boolean read FScriptLoaded;
  end;


implementation

{ TSimpleGate }

// if TSimpleGate is used at the same time, a Tag value must be unique
constructor TSimpleGate.Create(const Receiver: TObject; const WebBrowser: TWebBrowser; const Scheme: String; Tag: integer);
var
  SubscriptionId: Integer;
  MessageManager: TMessageManager;
begin
  FReceiver := Receiver;
  FWebBrowser := WebBrowser;
  //{$IFDEF MSWINDOWS}
  //FWebBrowser.WindowsEngine := TWindowsEngine.EdgeIfAvailable;
  //{$ENDIF}
  FScheme := LowerCase(Scheme);
  FTag := Tag;

  FScriptLoaded := false;

  FWebBrowser.OnShouldStartLoadWithRequest := WebBrowserOnShouldStartLoadWithRequest;
  FResultProc := nil;

  MessageManager := TMessageManager.DefaultManager;
  SubscriptionId := MessageManager.SubscribeToMessage(TMessage<UnicodeString>, MessageReceiver);
end;

procedure TSimpleGate.MessageReceiver(const Sender: TObject; const M: TMessage);
var
  AText: string;
begin
  AText := (M as TMessage<UnicodeString>).Value;
  if SameText(AText, 'ScriptLoaded_' + FTag.ToString) then begin
    FScriptLoaded := true;
  end else if AText.EndsWith('_' + FTag.ToString) then begin
    AText := AText.Remove(Length(AText) - 2);
    CallEvent(AText);
  end;
end;

function TSimpleGate.CheckURL(var AText: string): boolean;
begin
  result := false;

  AText := TIDURI.URLDecode(AText);
  if not AText.StartsWith('delphi:', true) then exit;
  if (SecondsBetween(FLastReturnTime, Now) < 2) and SameText(FLastReturnString, AText) then exit;

  FLastReturnString := AText;
  FLastReturnTime := now;

  AText := AText.Substring(FScheme.Length + 1);
  result := true;
end;

procedure TSimpleGate.WebBrowserOnShouldStartLoadWithRequest(ASender: TObject; const URL: string);
var
  AUrl: string;
  MessageManager: TMessageManager;
  Message: TMessage;
begin
  AUrl := URL;
  if CheckURL(AUrl) then begin
    if AUrl.StartsWith('return=', true) then begin
      if Assigned(FResultProc) then begin
        FResultProc(AUrl.Substring(Length('return=')));
        FResultProc := nil;
      end;
    end else if not FScriptLoaded and AUrl.StartsWith('ScriptLoaded', true) then begin
      FScriptLoaded := true;
      MessageManager := TMessageManager.DefaultManager;
      Message := TMessage<UnicodeString>.Create('ScriptLoaded_' + FTag.ToString);
      MessageManager.SendMessage(ASender, Message, True);
    end else begin
      // CallEvent(AUrl);
      MessageManager := TMessageManager.DefaultManager;
      Message := TMessage<UnicodeString>.Create(AUrl + '_' + FTag.ToString);
      MessageManager.SendMessage(ASender, Message, True);
    end;
    FWebBrowser.Stop;
  end;
end;

// from SG.ScriptGate.pas by HOSOKAWA
procedure TSimpleGate.CallEvent(AText: string);
begin
  TThread.CreateAnonymousThread(procedure
    const
      SCHEME_SEP = ':';
      METHOD_START_SEP = '(';
      METHOD_END_SEP = ')';
      PARAM_SEP = ',';
      SQUOTE = '''';
      DQUOTE = '"';
      SEMICOLON = ';';
    var
      iURL: string;
      Factors: TArray<String>;
      FactorsLen: Integer;
      Method: String;
      i: Integer;
      RttiType: TRttiType;
      RttiMethod: TRttiMethod;
      RttiParams: TArray<TRttiParameter>;
      VarParams: TArray<TValue>;
      ParamsLen: Integer;
      Factor: String;
    begin
      try
        iURL := AText;
        Factors := iURL.Split([METHOD_START_SEP]);
        Method := Factors[0];

        iURL := iURL.Substring(Method.Length + 1);
        Factors :=
          iURL.Split(
            [PARAM_SEP, METHOD_END_SEP],
            TStringSplitOptions.ExcludeEmpty);

        RttiType := SharedContext.GetType(FReceiver.ClassType);
        if RttiType = nil then exit;

        RttiMethod := RttiType.GetMethod(Method);
        if RttiMethod = nil then exit;

        FactorsLen := Length(Factors);
        RttiParams := RttiMethod.GetParameters;
        ParamsLen := Length(RttiParams);

        if ParamsLen <> FactorsLen then exit;

        SetLength(VarParams, Length(RttiParams));
        for i := Low(RttiParams) to High(RttiParams) do begin
          Factor := Factors[i].TrimRight;

          if (Factor.EndsWith(SEMICOLON)) then
            Factor := Factor.Substring(0, Factor.Length - 1).TrimRight;

          if (Factor.EndsWith(METHOD_END_SEP)) then
            Factor := Factor.Substring(0, Factor.Length - 1);

          Factor := Factor.Trim;
          Factor := Factor.DeQuotedString(SQUOTE);
          Factor := Factor.DeQuotedString(DQUOTE);

          case (RttiParams[i].ParamType.TypeKind) of
            TTypeKind.tkInteger,
            TTypeKind.tkInt64,
            TTypeKind.tkClass,
            TTypeKind.tkClassRef,
            TTypeKind.tkInterface,
            TTypeKind.tkPointer:
              try
                VarParams[i] := Factor.ToInteger;
              except
                VarParams[i] := 0;
              end;

            TTypeKind.tkFloat:
              VarParams[i] := Factor.ToDouble;

            TTypeKind.tkEnumeration:
              try
                VarParams[i] := Factor.ToBoolean;
              except
                VarParams[i] := Factor;
              end;

          else
            VarParams[i] := Factor;
          end;
        end;

        TThread.Synchronize(
          TThread.CurrentThread,
          procedure
          begin
            try
              RttiMethod.Invoke(FReceiver, VarParams);
            except

            end;
          end
        );
      except

      end;
    end).Start;
end;

procedure TSimpleGate.CallScript(const Script: String; const ResultProc: TResultProc);
begin
  FResultProc := ResultProc;
  FWebBrowser.EvaluateJavaScript(Script);
end;

end.
